{$A+,O+,R-}
unit CipherU;

interface

uses
  CommonU, HashU;

type
  ICryptCipher  = interface (ICryptBase)
                ['{57F06A42-EB24-11D3-8BEE-00C0DFECF5B7}']
    function    GetBlockSize: LongInt;
    function    GetMaxKeySize: LongInt;

    procedure   Init (var pKey; pSize: LongInt; pIVector: Pointer);
    procedure   InitStr (const pKey: string; pHash: ICryptHash);
    procedure   Reset;

    procedure   EncryptECB (const pInBlock; var pOutBlock);
    procedure   DecryptECB (const pInBlock; var pOutBlock);
    procedure   EncryptCBC (const pInBlock; var pOutBlock; pSize: LongInt);
    procedure   DecryptCBC (const pInBlock; var pOutBlock; pSize: LongInt);
    procedure   EncryptCFB (const pInBlock; var pOutBlock; pSize: LongInt);
    procedure   DecryptCFB (const pInBlock; var pOutBlock; pSize: LongInt);

    property    BlockSize: LongInt read GetBlockSize;
    property    MaxKeySize: LongInt read GetMaxKeySize;
  end;

type
  TCryptCipher  = class (TCryptBase, ICryptBase, ICryptCipher)
  protected
    FBlockSize  : LongInt;
    FMaxKeySize : LongInt;
    FNullInt    : LongInt;

    function    GetBlockSize: LongInt; virtual;
    function    GetMaxKeySize: LongInt; virtual;

    procedure   Init (var pKey; pSize: LongInt; pIVVector: Pointer); virtual; abstract;
    procedure   InitStr (const pKey: string; pHash: ICryptHash); virtual;
    procedure   Reset; virtual; abstract;

    procedure   EncryptECB (const pInBlock; var pOutBlock); virtual; abstract;
    procedure   DecryptECB (const pInBlock; var pOutBlock); virtual; abstract;
    procedure   EncryptCBC (const pInBlock; var pOutBlock; pSize: LongInt); virtual; abstract;
    procedure   DecryptCBC (const pInBlock; var pOutBlock; pSize: LongInt); virtual; abstract;
    procedure   EncryptCFB (const pInBlock; var pOutBlock; pSize: LongInt); virtual; abstract;
    procedure   DecryptCFB (const pInBlock; var pOutBlock; pSize: LongInt); virtual; abstract;

    property    BlockSize: LongInt read GetBlockSize;
    property    MaxKeySize: LongInt read GetMaxKeySize;

  public
    class       function IsHash: Boolean; override;
  end;

  function      CipherStr (pHash: ICryptHash; pCipher: ICryptCipher;
                    const pKey, pStr: string): string;

implementation

  function      CipherStr;
  var
    s           : string;

  begin
    with pCipher do
      begin
        SetLength (s, GetBlockSize div 8);

        InitStr (pKey, pHash);
        EncryptECB (pStr [1], s [1]);
      end;

    Result := s;
  end;
  
// TCryptCipher

  function      TCryptCipher.GetBlockSize;
  begin
    Result := FBlockSize
  end;

  function      TCryptCipher.GetMaxKeySize;
  begin
    Result := FMaxKeySize
  end;

  procedure     TCryptCipher.InitStr;
  var
    keyHash     : Pointer;

  begin
    pHash.Init;
    pHash.Update (pKey [1], Length (pKey));

    GetMem (keyHash, pHash.HashSize div 8);
    pHash.Final (keyHash^);

    if pHash.HashSize > MaxKeySize then
      Init (keyHash^, MaxKeySize, nil)
    else
      Init (keyHash^, pHash.HashSize, nil);

    FreeMem (keyHash, pHash.HashSize div 8);
    pHash.Burn;
  end;

  class         function TCryptCipher.IsHash;
  begin
    Result := False
  end;

end.

