{$A+,O+,R-}
unit CommonU;

interface

type
  TAlgID        = type LongInt;

  PByte         = ^Byte;
  PWord         = ^Word;
  PDWord        = ^DWord;
  DWord         = LongWord;

  PDWordArray   = ^TDWordArray;
  TDWordArray   = array [0 .. 1023] of DWord;
  PByteArray    = ^TByteArray;
  TByteArray    = array [0 .. 4095] of Byte;

{$IFDEF WIN32}
  function      LRot16 (X: Word; c: LongInt): Word; assembler;
  function      RRot16 (X: Word; c: LongInt): Word; assembler;
  function      LRot32 (X: DWord; c: LongInt): DWord; assembler;
  function      RRot32 (X: DWord; c: LongInt): DWord; assembler;
  function      SwapDWord (X: DWord): DWord; assembler;
{$ELSE}
  function      LRot16 (X: Word; c: LongInt): Word;
  function      RRot16 (X: Word; c: LongInt): Word;
  function      LRot32 (X: DWord; c: LongInt): DWord;
  function      RRot32 (X: DWord; c: LongInt): DWord;
  function      SwapDWord (X: DWord): DWord;
{$ENDIF}

  procedure     XorBlock (I1, I2, O1: PByteArray; Len: LongInt);

type
  ICryptBase    = interface (IUnknown)
                ['{57F06A40-EB24-11D3-8BEE-00C0DFECF5B7}']
    function    GetInitialized: Boolean;

    procedure   Burn;

    property    Initialized: Boolean read GetInitialized;
  end;

type
  TCryptBase    = class (TInterfacedObject, ICryptBase)
  protected
    FInitialized: Boolean;
    FNullStr    : string;

    function    GetInitialized: Boolean; virtual;

    procedure   Burn; virtual; abstract;

    property    Initialized: Boolean read GetInitialized;

  public
    constructor Create; virtual;
    destructor  Destroy; override;

    class       function IsHash: Boolean; virtual; abstract;

    class       function ID: TAlgID; virtual; abstract;
    class       function Algorithm: string; virtual; abstract;
  end;

// registration procedures

type
  TCryptClass   = class of TCryptBase;

  procedure     RegisterCrypt (pCrypt: TCryptClass);
  function      GetCrypt (pID: TAlgID): ICryptBase;

  function      GetCryptCount: Integer;
  function      GetCryptByIndex (pIndex: Integer): TCryptClass;

implementation

uses
  ContNrs;

var
  Algorithms    : TClassList;

{$IFDEF WIN32}
  function      LRot16 (X: Word; c: LongInt): Word; assembler;
  asm
    mov ecx,&c
    mov ax,&X
    rol ax,cl
    mov &Result,ax
  end;

  function      RRot16 (X: Word; c: LongInt): Word; assembler;
  asm
    mov ecx,&c
    mov ax,&X
    ror ax,cl
    mov &Result,ax
  end;

  function      LRot32 (X: DWord; c: LongInt): DWord; register; assembler;
  asm
    mov ecx, edx
    rol eax, cl
  end;

  function      RRot32 (X: DWord; c: LongInt): DWord; register; assembler;
  asm
    mov ecx, edx
    ror eax, cl
  end;

  function      SwapDWord (X: DWord): DWord; register; assembler;
  asm
    xchg al,ah
    rol  eax,16
    xchg al,ah
  end;

{$ELSE}
  function      LRot16 (X: Word; c: LongInt): Word;
  begin
    LRot16:= (X shl c) or (X shr (16 - c))
  end;

  function      RRot16 (X: Word; c: LongInt): Word;
  begin
    RRot16:= (X shr c) or (X shl (16 - c))
  end;

  function      LRot32 (X: DWord; c: LongInt): DWord;
  begin
    LRot32:= (X shl c) or (X shr (32 - c))
  end;

  function      RRot32 (X: DWord; c: LongInt): DWord;
  begin
    RRot32:= (X shr c) or (X shl (32 - c))
  end;

  function      SwapDWord (X: DWord): DWord;
  begin
    Result:= (X shr 24) or ((X shr 8) and $FF00) or ((X shl 8) and $FF0000) or (X shl 24)
  end;
{$ENDIF}

  procedure     XorBlock (I1, I2, O1: PByteArray; Len: LongInt);
  var
    i           : LongInt;

  begin
    for i := 0 to Len - 1 do
      O1^[i]:= I1^[i] xor I2^[i]
  end;

// TCryptBase

  function      TCryptBase.GetInitialized;
  begin
    Result := FInitialized
  end;

  constructor   TCryptBase.Create;
  begin
    inherited Create;

    Burn;
  end;

  destructor    TCryptBase.Destroy;
  begin
    if Initialized then
      Burn;

    inherited Destroy;
  end;

// registration procedures

  procedure     RegisterCrypt;
  begin
    Algorithms.Add (pCrypt)
  end;

  function      GetCrypt;
  var
    i           : Integer;
    cls         : TCryptClass;

  begin
    Result := nil;

    with Algorithms do
      for i := 0 to Pred (Count) do
        begin
          cls := TCryptClass (Items [i]);

          if cls.ID = pID then
            begin
              Result := cls.Create as ICryptBase;

              Exit;
            end;
        end;
  end;

  function      GetCryptCount;
  begin
    Result := Algorithms.Count
  end;

  function      GetCryptByIndex;
  begin
    Result := TCryptClass (Algorithms.Items [pIndex])
  end;

initialization
  Algorithms := TClassList.Create;

finalization
  Algorithms.Free;

end.

