{$A+,O+,R-}
unit HashU;

interface

uses
  CommonU;

type
  ICryptHash    = interface (ICryptBase)
                ['{57F06A41-EB24-11D3-8BEE-00C0DFECF5B7}']
    function    GetHashSize: LongInt;
    procedure   SetHashSize (pValue: LongInt);

    procedure   Init;
    procedure   Update (const pBuffer; pSize: LongInt);
    procedure   UpdateStr (const pBuffer: string);
    procedure   Final (var pDigest);
    function    FinalStr: string;

    property    HashSize: LongInt read GetHashSize write SetHashSize;
  end;

type
  TCryptHash    = class (TCryptBase, ICryptBase, ICryptHash)
  protected
    FHashSize   : LongInt;

    function    GetHashSize: LongInt; virtual;
    procedure   SetHashSize (pValue: LongInt); virtual;

    procedure   Init; virtual; abstract;
    procedure   Update (const pBuffer; pSize: LongInt); virtual; abstract;
    procedure   UpdateStr (const pBuffer: string);
    procedure   Final (var pDigest); virtual; abstract;
    function    FinalStr: string; virtual;

    property    HashSize: LongInt read GetHashSize write SetHashSize;

  public
    class       function IsHash: Boolean; override;
  end;

  function      HashStr (pHash: ICryptHash; const pStr: string): string;

implementation

  function      HashStr;
  begin
    Result := '';

    if Assigned (pHash) then
      with pHash do
        begin
          Init;
          UpdateStr (pStr);
          Result := FinalStr;
        end;
  end;

// TCryptHash

  function      TCryptHash.GetHashSize;
  begin
    Result := FHashSize
  end;

  procedure     TCryptHash.SetHashSize;
  begin
    // ...
  end;

  procedure     TCryptHash.UpdateStr;
  begin
    Update (pBuffer [1], Length (pBuffer))
  end;

  function      TCryptHash.FinalStr;
  var
    hashSize    : Integer;
    temp        : Pointer;
    s           : string;

  begin
    hashSize := GetHashSize div 8;
    GetMem (temp, hashSize);
    try
      Final (temp^);

      SetLength (s, hashSize);
      Move (temp^, s [1], hashSize);

    finally
      FreeMem (temp, hashSize);
    end;

    Result := s;
  end;

  class         function TCryptHash.IsHash;
  begin
    Result := True
  end;

end.

