{$A+,O+,R-}
unit SHA_1U;

interface

uses
  CommonU, HashU;

type
  TCryptSHA_1   = class (TCryptHash, ICryptBase, ICryptHash)
  protected
    LenHi,
    LenLo       : DWord;
    Index       : DWord;
    CurrentHash : array [0 ..  4] of DWord;
    HashBuffer  : array [0 .. 63] of Byte;

    procedure   Compress;
    procedure   UpdateLen (pLen: DWord);

  protected
    procedure   Burn; override;
    procedure   Init; override;

    procedure   Update (const pBuffer; pSize: LongInt); override;
    procedure   Final (var pDigest); override;

  public
    constructor Create; override;

    class       function ID: TAlgID; override;
    class       function Algorithm: string; override;
  end;

implementation

uses
  SysUtils;

// TCryptSHA_1

  procedure     TCryptSHA_1.Compress;
  var
    A, B, C, D,
    E, T        : DWord;
    W           : array [0 .. 79] of DWord;
    i           : LongInt;

  begin
    Index := 0;

    Move (HashBuffer, W, SizeOf (HashBuffer));
    for i := 0 to 15 do
      W [i] := (W [i] shr 24) or ((W [i] shr 8) and $FF00) or ((W [i] shl 8) and $FF0000) or (W [i] shl 24);
    for i := 16 to 79 do
      W [i] := LRot32 (W [i - 3] xor W [i - 8] xor W [i - 14] xor W [i - 16], 1);

    A := CurrentHash [0];
    B := CurrentHash [1];
    C := CurrentHash [2];
    D := CurrentHash [3];
    E := CurrentHash [4];

    for i := 0 to 19 do
      begin
        T := LRot32 (A, 5) + (D xor (B and (C xor D))) + E + W [i] + $5A827999;
        E := D; D := C; C := LRot32 (B, 30); B := A; A := T;
      end;
    for i := 20 to 39 do
      begin
        T := LRot32 (A, 5) + (B xor C xor D) + E + W [i] + $6ED9EBA1;
        E := D; D := C; C := LRot32 (B, 30); B := A; A := T;
      end;
    for i := 40 to 59 do
      begin
        T := LRot32 (A, 5) + ((B and C) or (D and (B or C))) + E + W [i] + $8F1BBCDC;
        E := D; D := C; C := LRot32 (B, 30); B := A; A := T;
      end;
    for i := 60 to 79 do
      begin
        T := LRot32 (A, 5) + (B xor C xor D) + E + W [i] + $CA62C1D6;
        E := D; D := C; C := LRot32 (B, 30); B := A; A := T;
      end;

    CurrentHash [0]:= CurrentHash [0] + A;
    CurrentHash [1]:= CurrentHash [1] + B;
    CurrentHash [2]:= CurrentHash [2] + C;
    CurrentHash [3]:= CurrentHash [3] + D;
    CurrentHash [4]:= CurrentHash [4] + E;

    FillChar (W, SizeOf (W), 0);
    FillChar (HashBuffer, SizeOf (HashBuffer), 0);
  end;

  procedure     TCryptSHA_1.UpdateLen;
  begin
    Inc (LenLo, (pLen shl 3));
    if LenLo < (pLen shl 3) then
      Inc (LenHi);
    Inc (LenHi, pLen shr 29);
  end;

  procedure     TCryptSHA_1.Burn;
  begin
    LenHi := 0;
    LenLo := 0;
    Index := 0;

    FillChar (HashBuffer, SizeOf (HashBuffer), 0);
    FillChar (CurrentHash, SizeOf (CurrentHash), 0);

    FInitialized:= false;
  end;

  procedure     TCryptSHA_1.Init;
  begin
    Burn;

    CurrentHash [0]:= $67452301;
    CurrentHash [1]:= $EFCDAB89;
    CurrentHash [2]:= $98BADCFE;
    CurrentHash [3]:= $10325476;
    CurrentHash [4]:= $C3D2E1F0;

    FInitialized:= True;
  end;

  procedure     TCryptSHA_1.Update;
  var
    PBuf        : PByte;

  begin
    if not Initialized then
      raise Exception.Create ('SHA_1: Not initialized');

    UpdateLen (pSize);

    PBuf:= @pBuffer;
    while pSize > 0 do
      begin
        if (SizeOf (HashBuffer) - Index) <= DWord (pSize) then
          begin
            Move (PBuf^, HashBuffer [Index], SizeOf (HashBuffer) - Index);
            Dec (pSize, SizeOf (HashBuffer) - Index);
            Inc (PBuf, SizeOf (HashBuffer) - Index);

            Compress;
          end
        else
          begin
            Move (PBuf^, HashBuffer [Index], pSize);
            Inc (Index, pSize);
            pSize := 0;
          end;
      end;
  end;

  procedure     TCryptSHA_1.Final;
  begin
    if not Initialized then
      raise Exception.Create ('SHA_1: Not initialized');

    HashBuffer [Index]:= $80;

    if Index >= 56 then
      Compress;

    PDWord (@HashBuffer [56])^ := (LenHi shr 24) or ((LenHi shr 8) and $FF00) or ((LenHi shl 8) and $FF0000) or (LenHi shl 24);
    PDWord (@HashBuffer [60])^ := (LenLo shr 24) or ((LenLo shr 8) and $FF00) or ((LenLo shl 8) and $FF0000) or (LenLo shl 24);

    Compress;

    CurrentHash [0]:= (CurrentHash [0] shr 24) or ((CurrentHash [0] shr 8) and $FF00)
      or ((CurrentHash [0] shl 8) and $FF0000) or (CurrentHash [0] shl 24);
    CurrentHash [1]:= (CurrentHash [1] shr 24) or ((CurrentHash [1] shr 8) and $FF00)
      or ((CurrentHash [1] shl 8) and $FF0000) or (CurrentHash [1] shl 24);
    CurrentHash [2]:= (CurrentHash [2] shr 24) or ((CurrentHash [2] shr 8) and $FF00)
      or ((CurrentHash [2] shl 8) and $FF0000) or (CurrentHash [2] shl 24);
    CurrentHash [3]:= (CurrentHash [3] shr 24) or ((CurrentHash [3] shr 8) and $FF00)
      or ((CurrentHash [3] shl 8) and $FF0000) or (CurrentHash [3] shl 24);
    CurrentHash [4]:= (CurrentHash [4] shr 24) or ((CurrentHash [4] shr 8) and $FF00)
      or ((CurrentHash [4] shl 8) and $FF0000) or (CurrentHash [4] shl 24);

    Move (CurrentHash, pDigest, SizeOf (CurrentHash));

    Burn;
  end;

  constructor   TCryptSHA_1.Create;
  begin
    inherited Create;

    FHashSize := 160;
  end;

  class         function TCryptSHA_1.ID;
  begin
    Result := 2
  end;

  class         function TCryptSHA_1.Algorithm;
  begin
    Result := 'SHA1'
  end;

initialization
  RegisterCrypt (TCryptSHA_1);

end.

